@@ -61,6 +61,7 @@ AVRO_TYPES = {
 
 decimal_context = Context()
 epoch = datetime(1970, 1, 1, tzinfo=timezone.utc)
+epoch_naive = datetime(1970, 1, 1)
 
 
 def match_types(writer_type, reader_type):
@@ -165,11 +166,25 @@ def read_timestamp_millis(data, writer_schema=None, reader_schema=None):
     return epoch + timedelta(microseconds=data * 1000)
 
 
+def read_local_timestamp_millis(
+    data: int, writer_schema=None, reader_schema=None
+) -> datetime:
+    # Cannot use datetime.fromtimestamp: https://bugs.python.org/issue36439
+    return epoch_naive + timedelta(microseconds=data * 1000)
+
+
 def read_timestamp_micros(data, writer_schema=None, reader_schema=None):
     # Cannot use datetime.fromtimestamp: https://bugs.python.org/issue36439
     return epoch + timedelta(microseconds=data)
 
 
+def read_local_timestamp_micros(
+    data: int, writer_schema=None, reader_schema=None
+) -> datetime:
+    # Cannot use datetime.fromtimestamp: https://bugs.python.org/issue36439
+    return epoch_naive + timedelta(microseconds=data)
+
+
 def read_date(data, writer_schema=None, reader_schema=None):
     return date.fromordinal(data + DAYS_SHIFT)
 
@@ -571,7 +586,9 @@ def skip_record(decoder, writer_schema, named_schemas):
 
 LOGICAL_READERS = {
     "long-timestamp-millis": read_timestamp_millis,
+    "long-local-timestamp-millis": read_local_timestamp_millis,
     "long-timestamp-micros": read_timestamp_micros,
+    "long-local-timestamp-micros": read_local_timestamp_micros,
     "int-date": read_date,
     "bytes-decimal": read_decimal,
     "fixed-decimal": read_decimal,
@@ -930,9 +947,7 @@ class file_reader:
         self._named_schemas = {}
         if reader_schema:
             self.reader_schema = parse_schema(
-                reader_schema,
-                _write_hint=False,
-                _named_schemas=self._named_schemas,
+                reader_schema, self._named_schemas, _write_hint=False
             )
         else:
             self.reader_schema = None
@@ -960,10 +975,7 @@ class file_reader:
         # Always parse the writer schema since it might have named types that
         # need to be stored in self._named_types
         self.writer_schema = parse_schema(
-            self._schema,
-            _write_hint=False,
-            _force=True,
-            _named_schemas=self._named_schemas,
+            self._schema, self._named_schemas, _write_hint=False, _force=True
         )
 
     @property
@@ -1163,7 +1175,7 @@ def schemaless_reader(fo, writer_schema, reader_schema=None, return_record_name=
         reader_schema = None
 
     named_schemas = {}
-    writer_schema = parse_schema(writer_schema, _named_schemas=named_schemas)
+    writer_schema = parse_schema(writer_schema, named_schemas)
 
     if reader_schema:
         reader_schema = parse_schema(reader_schema)
