@@ -1,21 +1,31 @@
 # cython: auto_cpdef=True
+import hashlib
+from io import StringIO
 import math
 from os import path
 from copy import deepcopy
 import json
 
 from ._schema_common import (
-    PRIMITIVES, UnknownType, SchemaParseException, RESERVED_PROPERTIES,
-    OPTIONAL_FIELD_PROPERTIES, RESERVED_FIELD_PROPERTIES,
+    PRIMITIVES,
+    UnknownType,
+    SchemaParseException,
+    RESERVED_PROPERTIES,
+    OPTIONAL_FIELD_PROPERTIES,
+    RESERVED_FIELD_PROPERTIES,
+    JAVA_FINGERPRINT_MAPPING,
+    FINGERPRINT_ALGORITHMS,
+    RABIN_64,
+    rabin_fingerprint,
 )
 
 
 def extract_record_type(schema):
     if isinstance(schema, dict):
-        return schema['type']
+        return schema["type"]
 
     if isinstance(schema, list):
-        return 'union'
+        return "union"
 
     return schema
 
@@ -24,11 +34,11 @@ def extract_logical_type(schema):
     if not isinstance(schema, dict):
         return None
     d_schema = schema
-    rt = d_schema['type']
-    lt = d_schema.get('logicalType')
+    rt = d_schema["type"]
+    lt = d_schema.get("logicalType")
     if lt:
         # TODO: Building this string every time is going to be relatively slow.
-        return '{}-{}'.format(rt, lt)
+        return f"{rt}-{lt}"
     return None
 
 
@@ -65,19 +75,19 @@ def fullname(schema):
 
 def schema_name(schema, parent_ns):
     try:
-        name = schema['name']
+        name = schema["name"]
     except KeyError:
-        msg = (
-            '"name" is a required field missing from '
-            + 'the schema: {}'.format(schema)
+        raise SchemaParseException(
+            f'"name" is a required field missing from the schema: {schema}'
         )
-        raise SchemaParseException(msg)
 
-    namespace = schema.get('namespace', parent_ns)
+    namespace = schema.get("namespace", parent_ns)
     if not namespace:
         return namespace, name
-
-    return namespace, '{}.{}'.format(namespace, name)
+    elif "." in name:
+        return "", name
+    else:
+        return namespace, f"{namespace}.{name}"
 
 
 def expand_schema(schema):
@@ -149,7 +159,12 @@ def expand_schema(schema):
 
 
 def parse_schema(
-    schema, expand=False, _write_hint=True, _force=False, _named_schemas=None,
+    schema,
+    named_schemas=None,
+    *,
+    expand=False,
+    _write_hint=True,
+    _force=False,
 ):
     """Returns a parsed avro schema
 
@@ -161,10 +176,9 @@ def parse_schema(
     ----------
     schema: dict
         Input schema
+    named_schemas: dict
+        Dictionary of named schemas to their schema definition
     expand: bool
-        NOTE: This option should be considered a keyword only argument and may
-        get enforced as such when Python 2 support is dropped.
-
         If true, named schemas will be fully expanded to their true schemas
         rather than being represented as just the name. This format should be
         considered an output only and not passed in to other reader/writer
@@ -176,9 +190,6 @@ def parse_schema(
     _force: bool
         Internal API argument. If True, the schema will always be parsed even
         if it has been parsed and has the __fastavro_parsed marker
-    _named_schemas: dict
-        Internal API argument. Dictionary of named schemas to their schema
-        definition
 
 
     Example::
@@ -189,40 +200,51 @@ def parse_schema(
         parsed_schema = parse_schema(original_schema)
         with open('weather.avro', 'wb') as out:
             writer(out, parsed_schema, records)
+
+
+    Sometimes you might have two schemas where one schema references another.
+    For the sake of example, let's assume you have a `Parent` schema that
+    references a `Child` schema`. If you were to try to parse the parent schema
+    on its own, you would get an exception because the child schema isn't
+    defined. To accomodate this, we can use the `named_schemas` argument to pass
+    a shared dictionary when parsing both of the schemas. The dictionary will
+    get populated with the necessary schema references to make parsing possible.
+    For example::
+
+        from fastavro import parse_schema
+
+        named_schemas = {}
+        parsed_child = parse_schema(child_schema, named_schemas)
+        parsed_parent = parse_schema(parent_schema, named_schemas)
     """
-    if _named_schemas is None:
-        _named_schemas = {}
+    if named_schemas is None:
+        named_schemas = {}
 
     if _force or expand:
-        return _parse_schema(
-            schema, "", expand, _write_hint, set(), _named_schemas
-        )
+        return _parse_schema(schema, "", expand, _write_hint, set(), named_schemas)
     elif isinstance(schema, dict) and "__fastavro_parsed" in schema:
         for key, value in schema["__named_schemas"].items():
-            _named_schemas[key] = value
+            named_schemas[key] = value
         return schema
     elif isinstance(schema, list):
         # If we are given a list we should make sure that the immediate sub
         # schemas have the hint in them
         return [
-            parse_schema(s, expand, _write_hint, _force, _named_schemas)
+            parse_schema(
+                s, named_schemas, expand=expand, _write_hint=_write_hint, _force=_force
+            )
             for s in schema
         ]
     else:
-        return _parse_schema(
-            schema, "", expand, _write_hint, set(), _named_schemas
-        )
+        return _parse_schema(schema, "", expand, _write_hint, set(), named_schemas)
 
 
-def _parse_schema(
-    schema, namespace, expand, _write_hint, names, named_schemas
-):
+def _parse_schema(schema, namespace, expand, _write_hint, names, named_schemas):
     # union schemas
     if isinstance(schema, list):
         return [
-            _parse_schema(
-                s, namespace, expand, False, names, named_schemas
-            ) for s in schema
+            _parse_schema(s, namespace, expand, False, names, named_schemas)
+            for s in schema
         ]
 
     # string schemas; this could be either a named schema or a primitive type
@@ -230,8 +252,8 @@ def _parse_schema(
         if schema in PRIMITIVES:
             return schema
 
-        if '.' not in schema and namespace:
-            schema = namespace + '.' + schema
+        if "." not in schema and namespace:
+            schema = namespace + "." + schema
 
         if schema not in named_schemas:
             raise UnknownType(schema)
@@ -268,8 +290,7 @@ def _parse_schema(
             scale = parsed_schema.get("scale")
             if scale and (not isinstance(scale, int) or scale < 0):
                 raise SchemaParseException(
-                    "decimal scale must be a postive integer, "
-                    + "not {}".format(scale)
+                    f"decimal scale must be a postive integer, not {scale}"
                 )
 
             precision = parsed_schema.get("precision")
@@ -277,25 +298,22 @@ def _parse_schema(
                 if not isinstance(precision, int) or precision <= 0:
                     raise SchemaParseException(
                         "decimal precision must be a postive integer, "
-                        + "not {}".format(precision)
+                        + f"not {precision}"
                     )
                 if schema_type == "fixed":
                     # https://avro.apache.org/docs/current/spec.html#Decimal
                     size = schema["size"]
-                    max_precision = int(
-                        math.floor(math.log10(2) * (8 * size - 1))
-                    )
+                    max_precision = int(math.floor(math.log10(2) * (8 * size - 1)))
                     if precision > max_precision:
-                        msg = "decimal precision of {} doesn't fit into " \
-                              "array of length {}"
                         raise SchemaParseException(
-                            msg.format(precision, size)
+                            f"decimal precision of {precision} doesn't fit "
+                            + f"into array of length {size}"
                         )
 
             if scale and precision and precision < scale:
                 raise SchemaParseException(
                     "decimal scale must be less than or equal to "
-                    + "the precision of {}".format(precision)
+                    + f"the precision of {precision}"
                 )
 
         if schema_type == "array":
@@ -321,9 +339,7 @@ def _parse_schema(
         elif schema_type == "enum":
             _, fullname = schema_name(schema, namespace)
             if fullname in names:
-                raise SchemaParseException(
-                    "redefined named type: {}".format(fullname)
-                )
+                raise SchemaParseException(f"redefined named type: {fullname}")
             names.add(fullname)
 
             named_schemas[fullname] = parsed_schema
@@ -334,9 +350,7 @@ def _parse_schema(
         elif schema_type == "fixed":
             _, fullname = schema_name(schema, namespace)
             if fullname in names:
-                raise SchemaParseException(
-                    "redefined named type: {}".format(fullname)
-                )
+                raise SchemaParseException(f"redefined named type: {fullname}")
             names.add(fullname)
 
             named_schemas[fullname] = parsed_schema
@@ -348,15 +362,13 @@ def _parse_schema(
             # records
             namespace, fullname = schema_name(schema, namespace)
             if fullname in names:
-                raise SchemaParseException(
-                    "redefined named type: {}".format(fullname)
-                )
+                raise SchemaParseException(f"redefined named type: {fullname}")
             names.add(fullname)
 
             named_schemas[fullname] = parsed_schema
 
             fields = []
-            for field in schema.get('fields', []):
+            for field in schema.get("fields", []):
                 fields.append(
                     parse_field(field, namespace, expand, names, named_schemas)
                 )
@@ -369,9 +381,7 @@ def _parse_schema(
                 # Make a copy of parsed_schema so that we don't have a cyclical
                 # reference. Using deepcopy is pretty slow, and we don't need a
                 # true deepcopy so this works good enough
-                named_schemas[fullname] = {
-                    k: v for k, v in parsed_schema.items()
-                }
+                named_schemas[fullname] = {k: v for k, v in parsed_schema.items()}
 
                 parsed_schema["__fastavro_parsed"] = True
                 parsed_schema["__named_schemas"] = named_schemas
@@ -399,8 +409,7 @@ def parse_field(field, namespace, expand, names, named_schemas):
     # Aliases must be a list
     aliases = parsed_field.get("aliases", [])
     if not isinstance(aliases, list):
-        msg = "aliases must be a list, not {}".format(aliases)
-        raise SchemaParseException(msg)
+        raise SchemaParseException(f"aliases must be a list, not {aliases}")
 
     parsed_field["name"] = field["name"]
     parsed_field["type"] = _parse_schema(
@@ -410,7 +419,9 @@ def parse_field(field, namespace, expand, names, named_schemas):
     return parsed_field
 
 
-def load_schema(schema_path, *, _named_schemas=None):
+def load_schema(
+    schema_path, *, named_schemas=None, _write_hint=True, _injected_schemas=None
+):
     """Returns a schema loaded from the file at `schema_path`.
 
     Will recursively load referenced schemas assuming they can be found in
@@ -421,9 +432,13 @@ def load_schema(schema_path, *, _named_schemas=None):
     ----------
     schema: str
         Path to schema file to load
-    _named_schemas: dict
-        Internal API argument. Dictionary of named schemas to their schema
-        definition
+    named_schemas: dict
+        Dictionary of named schemas to their schema definition
+    _write_hint: bool
+        Internal API argument specifying whether or not the __fastavro_parsed
+        marker should be added to the schema
+    _injected_schemas: set
+        Internal API argument. Set of names that have been injected
 
 
     Consider the following example...
@@ -434,6 +449,7 @@ def load_schema(schema_path, *, _named_schemas=None):
         {
             "type": "record",
             "name": "Parent",
+            "namespace": "namespace",
             "fields": [
                 {
                     "name": "child",
@@ -459,41 +475,298 @@ def load_schema(schema_path, *, _named_schemas=None):
 
         parsed_schema = load_schema("Parent.avsc")
     """
-    if _named_schemas is None:
-        _named_schemas = {}
+    if named_schemas is None:
+        named_schemas = {}
+
+    if _injected_schemas is None:
+        _injected_schemas = set()
 
     with open(schema_path) as fd:
         schema = json.load(fd)
     schema_dir, schema_file = path.split(schema_path)
-    return _load_schema(schema, schema_dir, _named_schemas)
+    return _load_schema(
+        schema, schema_dir, named_schemas, _write_hint, _injected_schemas
+    )
 
 
-def _load_schema(schema, schema_dir, named_schemas):
+def _load_schema(schema, schema_dir, named_schemas, write_hint, injected_schemas):
     try:
         schema_copy = deepcopy(named_schemas)
-        return parse_schema(schema, _named_schemas=named_schemas)
+        return parse_schema(schema, named_schemas=named_schemas, _write_hint=write_hint)
     except UnknownType as e:
         try:
-            avsc = path.join(schema_dir, '%s.avsc' % e.name)
-            sub_schema = load_schema(avsc, _named_schemas=schema_copy)
+            avsc = path.join(schema_dir, f"{e.name}.avsc")
+            sub_schema = load_schema(
+                avsc,
+                named_schemas=schema_copy,
+                _write_hint=False,
+                _injected_schemas=injected_schemas,
+            )
         except IOError:
             raise e
 
-        if isinstance(schema, dict):
-            if isinstance(sub_schema, list):
-                return _load_schema(
-                    sub_schema + [schema], schema_dir, schema_copy
-                )
+        if sub_schema["name"] not in injected_schemas:
+            injected_schema = _inject_schema(schema, sub_schema)
+            if isinstance(schema, str) or isinstance(schema, list):
+                schema = injected_schema[0]
+            injected_schemas.add(sub_schema["name"])
+        return _load_schema(
+            schema, schema_dir, schema_copy, write_hint, injected_schemas
+        )
+
+
+def _inject_schema(outer_schema, inner_schema, namespace="", is_injected=False):
+    # Once injected, we can stop checking to see if we need to inject since it
+    # should only be done once at most
+    if is_injected is True:
+        return outer_schema, is_injected
+
+    # union schemas
+    if isinstance(outer_schema, list):
+        union = []
+        for each_schema in outer_schema:
+            if is_injected:
+                union.append(each_schema)
             else:
-                return _load_schema(
-                    [sub_schema, schema], schema_dir, schema_copy
+                return_schema, injected = _inject_schema(
+                    each_schema, inner_schema, namespace, is_injected
                 )
+                union.append(return_schema)
+                if injected is True:
+                    is_injected = injected
+        return union, is_injected
+
+    # string schemas; this could be either a named schema or a primitive type
+    elif not isinstance(outer_schema, dict):
+        if outer_schema in PRIMITIVES:
+            return outer_schema, is_injected
+
+        if "." not in outer_schema and namespace:
+            outer_schema = namespace + "." + outer_schema
+
+        if outer_schema == inner_schema["name"]:
+            return inner_schema, True
         else:
-            # schema is already a list
-            if isinstance(sub_schema, list):
-                return _load_schema(
-                    sub_schema + schema, schema_dir, schema_copy
-                )
-            else:
-                schema.insert(0, sub_schema)
-                return _load_schema(schema, schema_dir, schema_copy)
+            # Hit a named schema that has already been loaded previously. Return
+            # the outer_schema so we keep looking
+            return outer_schema, is_injected
+    else:
+        # Remaining valid schemas must be dict types
+        schema_type = outer_schema["type"]
+
+        if schema_type == "array":
+            return_schema, injected = _inject_schema(
+                outer_schema["items"], inner_schema, namespace, is_injected
+            )
+            outer_schema["items"] = return_schema
+            return outer_schema, injected
+
+        elif schema_type == "map":
+            return_schema, injected = _inject_schema(
+                outer_schema["values"], inner_schema, namespace, is_injected
+            )
+            outer_schema["values"] = return_schema
+            return outer_schema, injected
+
+        elif schema_type == "enum":
+            return outer_schema, is_injected
+
+        elif schema_type == "fixed":
+            return outer_schema, is_injected
+
+        elif schema_type == "record" or schema_type == "error":
+            # records
+            namespace, _ = schema_name(outer_schema, namespace)
+            fields = []
+            for field in outer_schema.get("fields", []):
+                if is_injected:
+                    fields.append(field)
+                else:
+                    return_schema, injected = _inject_schema(
+                        field["type"], inner_schema, namespace, is_injected
+                    )
+                    field["type"] = return_schema
+                    fields.append(field)
+
+                    if injected is True:
+                        is_injected = injected
+            if fields:
+                outer_schema["fields"] = fields
+
+            return outer_schema, is_injected
+
+        elif schema_type in PRIMITIVES:
+            return outer_schema, is_injected
+
+        else:
+            raise Exception(
+                "Internal error; "
+                + "You should raise an issue in the fastavro github repository"
+            )
+
+
+def load_schema_ordered(ordered_schemas, *, _write_hint=True):
+    """Returns a schema loaded from a list of schemas.
+
+    The list of schemas should be ordered such that any dependencies are listed
+    before any other schemas that use those dependencies. For example, if schema
+    `A` depends on schema `B` and schema B depends on schema `C`, then the list
+    of schemas should be [C, B, A].
+
+    Parameters
+    ----------
+    ordered_schemas: list
+        List of paths to schemas
+    _write_hint: bool
+        Internal API argument specifying whether or not the __fastavro_parsed
+        marker should be added to the schema
+
+
+    Consider the following example...
+
+
+    Parent.avsc::
+
+        {
+            "type": "record",
+            "name": "Parent",
+            "namespace": "namespace",
+            "fields": [
+                {
+                    "name": "child",
+                    "type": "Child"
+                }
+            ]
+        }
+
+
+    namespace.Child.avsc::
+
+        {
+            "type": "record",
+            "namespace": "namespace",
+            "name": "Child",
+            "fields": []
+        }
+
+
+    Code::
+
+        from fastavro.schema import load_schema_ordered
+
+        parsed_schema = load_schema_ordered(
+            ["path/to/namespace.Child.avsc", "path/to/Parent.avsc"]
+        )
+    """
+    loaded_schemas = []
+    named_schemas = {}
+    for idx, schema_path in enumerate(ordered_schemas):
+        # _write_hint is always False except maybe the outer most schema
+        _last = _write_hint if idx + 1 == len(ordered_schemas) else False
+        schema = load_schema(
+            schema_path, named_schemas=named_schemas, _write_hint=_last
+        )
+        loaded_schemas.append(schema)
+
+    top_first_order = loaded_schemas[::-1]
+    outer_schema = top_first_order.pop(0)
+
+    while top_first_order:
+        sub_schema = top_first_order.pop(0)
+        _inject_schema(outer_schema, sub_schema)
+
+    return outer_schema
+
+
+def to_parsing_canonical_form(schema):
+    """Returns a string represening the parsing canonical form of the schema.
+
+    For more details on the parsing canonical form, see here:
+    https://avro.apache.org/docs/current/spec.html#Parsing+Canonical+Form+for+Schemas
+    """
+    fo = StringIO()
+    _to_parsing_canonical_form(parse_schema(schema), fo)
+    return fo.getvalue()
+
+
+def _to_parsing_canonical_form(schema, fo):
+    # union schemas
+    if isinstance(schema, list):
+        fo.write("[")
+        for idx, s in enumerate(schema):
+            if idx != 0:
+                fo.write(",")
+            _to_parsing_canonical_form(s, fo)
+        fo.write("]")
+
+    # string schemas; this could be either a named schema or a primitive type
+    elif not isinstance(schema, dict):
+        fo.write(f'"{schema}"')
+
+    else:
+        # Remaining valid schemas must be dict types
+        schema_type = schema["type"]
+
+        if schema_type == "array":
+            fo.write(f'{{"type":"{schema_type}","items":')
+            _to_parsing_canonical_form(schema["items"], fo)
+            fo.write("}")
+
+        elif schema_type == "map":
+            fo.write(f'{{"type":"{schema_type}","values":')
+            _to_parsing_canonical_form(schema["values"], fo)
+            fo.write("}")
+
+        elif schema_type == "enum":
+            name = schema["name"]
+            fo.write(f'{{"name":"{name}","type":"{schema_type}","symbols":[')
+
+            for idx, symbol in enumerate(schema["symbols"]):
+                if idx != 0:
+                    fo.write(",")
+                fo.write(f'"{symbol}"')
+            fo.write("]}")
+
+        elif schema_type == "fixed":
+            name = schema["name"]
+            size = schema["size"]
+            fo.write(f'{{"name":"{name}","type":"{schema_type}","size":{size}}}')
+
+        elif schema_type == "record" or schema_type == "error":
+            name = schema["name"]
+            fo.write(f'{{"name":"{name}","type":"record","fields":[')
+
+            for idx, field in enumerate(schema["fields"]):
+                if idx != 0:
+                    fo.write(",")
+                name = field["name"]
+                fo.write(f'{{"name":"{name}","type":')
+                _to_parsing_canonical_form(field["type"], fo)
+                fo.write("}")
+            fo.write("]}")
+
+        elif schema_type in PRIMITIVES:
+            fo.write(f'"{schema_type}"')
+
+
+def fingerprint(parsing_canonical_form, algorithm):
+    """Returns a string represening a fingerprint/hash of the parsing canonical
+    form of a schema.
+
+    For more details on the fingerprint, see here:
+    https://avro.apache.org/docs/current/spec.html#schema_fingerprints
+    """
+    if algorithm not in FINGERPRINT_ALGORITHMS:
+        raise ValueError(
+            f"Unknown schema fingerprint algorithm {algorithm}. "
+            + f"Valid values include: {FINGERPRINT_ALGORITHMS}"
+        )
+
+    # Fix Java names
+    algorithm = JAVA_FINGERPRINT_MAPPING.get(algorithm, algorithm)
+
+    if algorithm == RABIN_64:
+        return rabin_fingerprint(parsing_canonical_form.encode())
+
+    h = hashlib.new(algorithm, parsing_canonical_form.encode())
+    return h.hexdigest()
